#include "VideoSource.h"
#include "Video.h"
#include <iomanip>
#include <file/Path.h>
#include <misc/GlobalSettings.h>
#include <misc/ThreadPool.h>

using namespace cmn;

std::vector<std::pair<std::string, VideoSource::File::Type>> VideoSource::File::_extensions = {
    { "mov", VIDEO },
    { "mp4", VIDEO },
    { "mxf", VIDEO },
    { "avi", VIDEO },
    { "mts", VIDEO },
    { "tiff", IMAGE },
    { "tif", IMAGE },
    { "jpg", IMAGE },
	{ "png", IMAGE },
    { "bmp", IMAGE }
};

std::string VideoSource::File::complete_name(const std::string &basename, const std::string &ext) {
    return basename + "." + ext;
}

VideoSource::File * VideoSource::File::open(long_t index, const std::string& basename, const std::string& ext, bool no_check) {
    if(no_check) {
        return new File(index, basename, ext);
    }
    
    if (file_exists(complete_name(basename, ext))) {
        return new File(index, basename, ext);
    } //else {
        //U_EXCEPTION("Cannot open file '%S.%S'.", &basename, &ext);
    //}
    
    return NULL;
}

VideoSource::File::File(long_t index, const std::string& basename, const std::string& extension) : _index(index), _video(NULL), _size(0, 0) {
    _filename = complete_name(basename, extension);
    
    // check the extension(-type)
    _type = UNKNOWN;
    for (auto ext : _extensions) {
        if (ext.first == utils::lowercase(extension)) {
            _type = ext.second;
            break;
        }
    }
    
    if (_type == UNKNOWN)
        U_EXCEPTION("Unknown extension '%S' for file '%S'.", &extension, &_filename);
    
    switch (_type) {
        case VIDEO: {
            _video = new Video();
            auto npz = file::Path(_filename).replace_extension("npz");
            if(npz.exists()) {
                static bool message = false;
                if(!message) {
                    Debug("Found timestamps for file '%S'.", &npz.str());
                    message = true;
                }
                
                _timestamps = cnpy::npz_load(npz.str(), "frame_time").as_vec<double>();
                auto res = cnpy::npz_load(npz.str(), "imgshape").as_vec<int64_t>();
                _size = cv::Size( (int)res[1], (int)res[0] );
                _length = (long_t)_timestamps.size();
                
            } else {
                if(!_video->open(_filename))
                    U_EXCEPTION("Opening Video '%S' failed.", &_filename);
                _length = _video->length();
                _video->close();
            }
            break;
        }
            
        case IMAGE:
            _length = 1;
            break;
            
        default:
            break;
    }
}

void VideoSource::File::frame(long_t frameIndex, cv::Mat& output, bool lazy_video) const {
    switch (_type) {
        case VIDEO:
            if (!_video->isOpened())
                _video->open(_filename);
            if (!_video->isOpened())
                U_EXCEPTION("Video '%S' cannot be opened.", &_filename);
            _video->frame(frameIndex, output, lazy_video);
            break;
            
        case IMAGE:
            cv::imread(_filename, cv::IMREAD_GRAYSCALE).copyTo(output);
            break;
            
        default:
            U_EXCEPTION("Grabbing frame %d from '%S' failed because the type was unknown.", frameIndex, &_filename);
    }
}

bool VideoSource::File::has_timestamps() const {
    switch (_type) {
        case VIDEO:
            /*if (!_video->isOpened())
                _video->open(_filename);
            
            output = _video->frame(frameIndex);*/
            return !_timestamps.empty();
            
        case IMAGE:
            return false;
            
        default:
            break;
    }
    
    U_EXCEPTION("Retrieving timestamp for '%S' failed because the type was unknown.", &_filename);
}

short VideoSource::File::framerate() {
    if(type() != VIDEO)
        return -1;
    
    if(has_timestamps()) {
        if(_timestamps.size() > 1) {
            auto prev = _timestamps[0];
            double average = 0;
            for (uint64_t i=1; i<_timestamps.size(); i++) {
                average += _timestamps[i] - prev;
                prev = _timestamps[i];
            }
            average = average / double(_timestamps.size()-1);
            return (short)round(1. / average);
        } else
            return -1;
        
    } else {
        if(!_video->isOpened())
            _video->open(_filename);
        auto fps = _video->framerate();
        _video->close();
        return fps;
    }
}

uint64_t VideoSource::File::timestamp(uint64_t frameIndex) const {
    if(_type != VIDEO)
        U_EXCEPTION("Cannot retrieve timestamps from anything else other than videos.");
    
    if(!has_timestamps())
        U_EXCEPTION("No timestamps available for '%S'.", &_filename);
    
    auto times = _timestamps[frameIndex];
    auto point = std::chrono::duration<double>(times);
    auto duration = std::chrono::duration_cast<std::chrono::nanoseconds>(point);
    //uint64_t seconds = std::chrono::duration_cast<std::chrono::seconds>(duration).count();
    //uint64_t ms = std::chrono::duration_cast<std::chrono::milliseconds>(duration).count();
    uint64_t ns = std::chrono::duration_cast<std::chrono::microseconds>(duration).count();
    return ns;
}

const cv::Size& VideoSource::File::resolution() {
    if(_size.width == 0 && _size.height == 0) {
        switch(_type) {
            case VIDEO:
                if(!_video->isOpened())
                    _video->open(_filename);
                _size = _video->size();
                break;
                
            case IMAGE: {
                auto output = cv::imread(_filename, cv::IMREAD_GRAYSCALE);
                _size = {output.cols, output.rows};
                break;
            }
            default:
                U_EXCEPTION("Retrieving size from '%S' failed because the type was unknown.", &_filename);
        }
    }
    
    return _size;
}

void VideoSource::File::close() const {
    if (_type == VIDEO) {
        _video->close();
    }
}

VideoSource::File::~File() {
    if (_video) {
        delete _video;
    }
}

VideoSource::~VideoSource() {
    for (auto info : _files_in_seq) {
        delete info;
    }
}

VideoSource::VideoSource(const std::vector<file::Path>& files) : _last_file(NULL), _length(0), _has_timestamps(false), _framerate(-1)
{
    for(auto &path : files) {
        auto extension = path.extension().to_string();
        auto basename = path.remove_extension().str();
        auto f = File::open(_files_in_seq.size(), basename, extension);
        if(!f)
            U_EXCEPTION("Cannot open file '%S'.", &path.str());
        
        _files_in_seq.push_back(f);
        _length += (long_t)f->length();
    }
    
    if(_files_in_seq.empty()) {
        auto str = Meta::toStr(files);
        U_EXCEPTION("Cannot load video sequence '%S' (it is empty).", &str);
    }
    
    _size = _files_in_seq.at(0)->resolution();
    _has_timestamps = _files_in_seq.front()->has_timestamps();
}

VideoSource::VideoSource(const std::string& basename, const std::string& extension, int seq_start, int seq_end, int padding) : _last_file(NULL), _length(0), _has_timestamps(false), _framerate(-1) {
    
    if (seq_start == VIDEO_SEQUENCE_INVALID_VALUE || seq_end == VIDEO_SEQUENCE_INVALID_VALUE) {
        File *f = File::open(0, basename, extension);

        if(f && f->type() != File::VIDEO)
            Warning("Just loading one image because seq_end/seq_start were not specified.");

        if(f) {
            _files_in_seq.push_back(f);
            _length += (long_t)f->length();
        } else {
            U_EXCEPTION("Input source '%S' not found.", &basename);
        }
        
    } else if(seq_end == VIDEO_SEQUENCE_UNSPECIFIED_VALUE) {
        auto base = file::Path(basename).is_folder() ? "" : file::Path(basename).filename().to_string();
        Debug("Trying to find the last file (starting at %d) pattern '%S%%%dd.%S'...", seq_start, &base, padding, &extension);
        
        _files_in_seq.reserve(10000);
        
        int i =seq_start;
        do {
            std::stringstream ss;
            
            try {
                ss << basename << std::setfill('0') << std::setw(padding) << i;
                File *file = File::open(i - seq_start, ss.str(), extension);
                if(!file) {
                    break;
                }
                
                _files_in_seq.push_back(file);
                
                _length += file->length();
                
                i++;
                
            } catch (const UtilsException& ex) {
                break;
            }
            
            if(i%10000 == 0)
                Debug("Finding file %d (%d found)...", i, _files_in_seq.size());
            if(SETTING(terminate))
                break;
            
        } while (true);
        
        Debug("Last number was %d", i-1);
        _files_in_seq.shrink_to_fit();
        
    } else {
        Debug("Finding all relevant files in sequence with base name '%S'...", &basename);
        for (int i=seq_start; i<=seq_end; i++) {
            std::stringstream ss;
            ss << basename << std::setfill('0') << std::setw(padding) << i;
            
            File *f = File::open(i-seq_start, ss.str(), extension, i != seq_start);
			if(!f)
				U_EXCEPTION("Cannot find file '%s.%S' in sequence %d-%d.", ss.str().c_str(), &extension, seq_start, seq_end);
            _files_in_seq.push_back(f);
            
            _length += f->length();
            
            if(type() != File::Type::IMAGE && i % 50 == 0)
                Debug("%d/%d", i, seq_end);
            
            if(SETTING(terminate))
                break;
        }
        
        if (_files_in_seq.empty())
            U_EXCEPTION("Provided an empty video sequence for video source '%S'.", &basename);
    }
    
	if(_files_in_seq.empty())
		U_EXCEPTION("Cannot load video sequence '%S' (it is empty).", &basename);
	
    _size = _files_in_seq.at(0)->resolution();
    _has_timestamps = _files_in_seq.front()->has_timestamps();
    
    /**
     * Because of a bug in loop bio software while saving specific videos,
     * they sometimes put slightly wrong metadata in the accompanying npz files.
     * In this case we have to disregard the dimensions stated within and use
     * the actual ones from the frame.
     **/
    if(_has_timestamps) {
        auto first = _files_in_seq.front();
        cv::Mat image;
        first->frame(0, image);
        if(image.cols != _size.width || image.rows != _size.height) {
            Warning("VideoSource '%S' reports resolution %dx%d in metadata, but is actually %dx%d. Going with the actual video dimensions for now.", &basename, _size.width, _size.height, image.cols, image.rows);
            _size = cv::Size(image.cols, image.rows);
        }
        _last_file = first;
    }
    
    Debug("Resolution of VideoSource '%S' is [%dx%d]", &basename, _size.width, _size.height);
    
    if(type() == File::VIDEO) {
        _framerate = _files_in_seq.at(0)->framerate();
    }
}

void VideoSource::frame(uint64_t globalIndex, gpuMat& output) {
}

void VideoSource::frame(uint64_t globalIndex, cv::Mat& output) {
    if (/*globalIndex < 0 ||*/ globalIndex >= _length)
        U_EXCEPTION("Invalid frame %d/%d requested.", globalIndex, _length);
    
    if(type() == File::Type::IMAGE) {
        auto f = _files_in_seq.at(globalIndex);
        if(_last_file && _last_file != f)
            _last_file->close();
        
        _last_file = f;
        f->frame(0, output);
        
        if(output.empty())
            U_EXCEPTION("Could not find frame %d/%d in VideoSource.", globalIndex, length());
        
        return;
        
    } else {
        uint64_t index = 0;
        
        for (auto f : _files_in_seq) {
            if (index + f->length() > globalIndex) {
                if (_last_file && _last_file != f) {
                    _last_file->close();
                }
                
                _last_file = f;
                f->frame(globalIndex-index, output);
                
                if(output.empty())
                    U_EXCEPTION("Could not find frame %d/%d in VideoSource.", globalIndex, length());
                
                return;
            }
            
            index += f->length();
        }
        
        U_EXCEPTION("Could not find frame %d/%d in VideoSource.", globalIndex, index);
    }
}

uint64_t VideoSource::timestamp(uint64_t globalIndex) const {
    if (/*globalIndex < 0 ||*/ globalIndex >= _length)
        U_EXCEPTION("Invalid frame %d/%d requested.", globalIndex, _length);
    
    uint64_t index = 0;
    
    for (auto f : _files_in_seq) {
        if (index + f->length() > globalIndex) {
            return f->timestamp(globalIndex - index);
        }
        
        index += f->length();
    }
    
    U_EXCEPTION("Could not find frame %d/%d in VideoSource.", globalIndex, index);
}

#include <locale>

uint64_t VideoSource::start_timestamp() const {
    return _files_in_seq.front()->timestamp(0);
}

bool VideoSource::has_timestamps() const {
    return _has_timestamps;
}

short VideoSource::framerate() const {
    return _framerate;
}

void VideoSource::generate_average(cv::Mat &av, uint64_t) {
    gpuMat average;
    av.copyTo(average);
    // if there are only a few files, we can use the standard method
    /*if(_files_in_seq.size() <= 50) {
        GenericVideo::generate_average(average, frameIndex);
        return;
    }*/
    
    // if there are many files, we should use only the first part of each video
    Debug("Generating multi-file average...");
    
    //const uint64_t channel = SETTING(color_channel);
    double count = 0;
    
    gpuMat float_mat, f, ref;
    std::vector<gpuMat> vec;
    
    AveragingMethod::Class method(AveragingMethod::mean);
    if(GlobalSettings::has("averaging_method")) {
        auto name = SETTING(averaging_method).value<std::string>();
        if(AveragingMethod::has(name))
            method = AveragingMethod::get(name);
        else {
            auto str = Meta::toStr(AveragingMethod::names);
            Warning("Invalid value for 'averaging_method': '%S'. Known methods are: %S", &name, &str);
        }
    }
    //bool use_mean = GlobalSettings::has("averaging_method") && utils::lowercase(SETTING(averaging_method).value<std::string>()) != "max";
    Debug("Use averaging method: '%s'", method.name());
    
    if(average.empty() || average.cols != size().width || average.rows != size().height) {
        average = gpuMat::zeros(size().height, size().width, method == AveragingMethod::mean ? CV_32FC1 : CV_8UC1);
    } else
        average = cv::Scalar(0);
    
    if (length() < 10) {
        processImage(average, average);
        return;
    }
    
    if(method == AveragingMethod::mean)
        float_mat = gpuMat::zeros(average.rows, average.cols, CV_32FC1);
    else {
        if(method == AveragingMethod::min) {
            cv::Mat tmp = cv::Mat::ones(average.rows, average.cols, CV_8UC1);
            tmp = tmp.mul(255);
            tmp.copyTo(float_mat);
            
        } else
            float_mat = gpuMat::zeros(average.rows, average.cols, CV_8UC1);
    }
    
    float samples = GlobalSettings::has("average_samples") ? SETTING(average_samples).value<int>() : (length() * 0.01);
    uint64_t step = max(1, _files_in_seq.size() < samples ? 1 : ceil(_files_in_seq.size() / samples));
    uint64_t frames_per_file = max(1, _files_in_seq.size() < samples ? (length() / _files_in_seq.size()) / (length() / samples) : 1);
    
    if(samples > 255 && method == AveragingMethod::mode)
        U_EXCEPTION("Cannot take more than 255 samples with 'averaging_method' = 'mode'. Choose fewer samples or a different averaging method.");
    std::map<File*, std::set<uint64_t>> file_indexes;
    
    //std::vector<std::map<uchar, uint8_t>> spatial_histogram;
    std::vector<std::array<uint8_t, 256>> spatial_histogram;
    std::vector<std::unique_ptr<std::mutex>> spatial_mutex;
    if(method == AveragingMethod::mode) {
        spatial_histogram.resize(average.cols * average.rows);
        for(uint64_t i=0; i<spatial_histogram.size(); ++i) {
            std::fill(spatial_histogram.at(i).begin(), spatial_histogram.at(i).end(), 0);
            spatial_mutex.push_back(std::make_unique<std::mutex>());
        }
    }
    
    Debug("generating average in threads step %d for %d files (%d per file)", step, _files_in_seq.size(), frames_per_file);
    
    std::mutex mutex;
    GenericThreadPool pool(cmn::hardware_concurrency(), [](auto e) { std::rethrow_exception(e); }, "AverageImage");
    for(uint64_t i=0; i<_files_in_seq.size(); i+=step) {
        auto file = _files_in_seq.at(i);
        file_indexes[file].insert(0);
        if(frames_per_file > 1) {
            auto step = max(1, ceil(file->length() / frames_per_file));
            for(uint64_t i=step; i<(uint64_t)file->length(); i+= step) {
                file_indexes[file].insert(i);
            }
        }
    }
    
    for(auto && [file, indexes] : file_indexes) {
        auto fn = [method, samples, gAverage = &average, gAv = &av, gCount = &count,&mutex, &spatial_histogram, &spatial_mutex, &file_indexes](File* file, const std::set<uint64_t>& indexes){
            if(SETTING(terminate))
                return;
            
            cv::Mat f;
            cv::Mat float_mat;
            cv::Mat average;
            
            if(method == AveragingMethod::min) {
                average = cv::Mat::ones(gAverage->rows, gAverage->cols, gAverage->type());
                cv::multiply(average, cv::Scalar(255), average);
            } else
                average = cv::Mat::zeros(gAverage->rows, gAverage->cols, gAverage->type());
            
            double count = 0;
            
            for(auto index : indexes) {
                try {
                    file->frame(index, f, true);
                    assert(f.channels() == 1);
                    
                    if(method == AveragingMethod::mean) {
                        //Debug("%d,%d - %d,%d", float_mat.cols, float_mat.rows, average.cols, average.rows);
                        f.convertTo(float_mat, CV_32FC1, 1.0/255.0);
                        cv::add(average, float_mat, average);
                        
                    } else if(method == AveragingMethod::mode) {
                        assert(f.isContinuous());
                        
                        auto ptr = f.data;
                        const auto end = f.data + f.cols * f.rows;
                        auto array_ptr = spatial_histogram.data();
                        auto mutex_ptr = spatial_mutex.begin();
                        
                        assert(spatial_histogram.size() == (uint64_t)(f.cols * f.rows));
                        if(file_indexes.size() > 1) {
                            for (; ptr != end; ++ptr, ++array_ptr, ++mutex_ptr) {
                                (*mutex_ptr)->lock();
                                ++(*array_ptr)[*ptr];
                                (*mutex_ptr)->unlock();
                            }
                        } else {
                            for (; ptr != end; ++ptr, ++array_ptr)
                                ++(*array_ptr)[*ptr];
                        }
                        
                    } else if(method == AveragingMethod::max) {
                        average = cv::max(average, f);
                    } else if(method == AveragingMethod::min) {
                        average = cv::min(average, f);
                    } else
                        U_EXCEPTION("Unknown averaging_method '%s'.", method.name())
                
                    ++count;
                    
                    if(long_t(count) % max(1,long_t(samples * 0.1)) == 0) {
                        Debug("%.0f / %.0f ('%S')", count, samples, &file->filename());
                    }
                    
                } catch (const UtilsException& e) {
                    Warning("Continuing, but caught an exception processing frame %d of '%S' while generating an average.", index, &file->filename());
                }
                
                if(SETTING(terminate))
                    return;
            }
            
            file->close();
            
            std::lock_guard<std::mutex> guard(mutex);
            if(method == AveragingMethod::mean) {
                *gCount += count;
                cv::add(average, *gAverage, *gAverage);
            } else if(method == AveragingMethod::mode) {
                
            } else {
                if(gAv->empty())
                    *gAv = average;
                else if(method == AveragingMethod::max)
                    *gAv = cv::max(average, *gAv);
                else
                    *gAv = cv::min(average, *gAv);
            }
        };
        
        if(file_indexes.size() > 1) {
            pool.enqueue(fn, file, indexes);
        } else
            fn(file, indexes);
    }
    
    pool.wait();
    _last_file = NULL;
    
    if(SETTING(terminate))
        return;
    
    if(method == AveragingMethod::mean) {
        cv::divide(average, cv::Scalar(count), average);
        average.convertTo(av, CV_8UC1, 255.0);
        
    } else if(method == AveragingMethod::mode) {
        Debug("Combining mode image...");
        average.copyTo(av);
        
        auto ptr = av.data;
        const auto end = av.data + av.cols * av.rows;
        auto array_ptr = spatial_histogram.data();
        
        for (; ptr != end; ++ptr, ++array_ptr) {
            uchar max_code = 0;
            uint8_t max_number = 0;
            //for(auto && [code, number] : *array_ptr) {
            for(uint64_t code=0; code<array_ptr->size(); ++code) {
                const auto& number = (*array_ptr)[code];
                if(number > max_number) {
                    max_number = number;
                    max_code = code;
                }
            }
            
            *ptr = max_code;
        }
    }
}
